// OrderGenerator.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;

/**
 * The OrderGenerator is used at present to simulate the market
 * behavior.
 *
 * @author Pietro Terna */
public class OrderGenerator  extends SwarmObjectImpl{

    /** How many unit we have, used at present to define the max value
	used to generate the 'digits' composing our recipes */

    /** the max length of a recipe */
    public int maxStepNumber;

    /** the max length of a step in a recipe */
    public int maxStepLength;

    /** used to record the number of generated orders*/
    public int orderCount;

    /** a specific order*/
    public Order anOrder;
    /** the list containig the operating units*/
    public ListImpl unitList;
   /** the list containig the end units*/
    public ListImpl endUnitList;
    /** the list containig all the orders*/
    public ListImpl orderList;
    /** the model (to which we apply mark up of the order revenue */
    VEFrameModelSwarm vEFrameModelSwarm;

    /** the integer vector containing a specific production recipe */
    public int[] orderRecipe1; 

    /** the integer vector containing the unit of time spent for each step 
     *  in the recipe; il will determine the tranformation of the orderRecipe1 
     *  vector into the orderRecipe2 vector containing the repetitions
     *  of the steps, to reproduce the length of each step in time
     *
     *  NB this formalism is internal to internal (in external formalism
     *  we have 'step s/h/d time'
     *
     *  as an examples
     * 
     *  orderRecipe1      contains  1 12  7  3
     *  stepLengthInOrder contains  1  3  2  2
     *
     *  orderRecipe2      contains  1 12 12 12  7  7  3  3 
     */
    public int[] stepLengthInOrder;

    /** the integer vector containing a specific production recipe,
     *  developed following the length of each step
     */
    public int[] orderRecipe2;

    /** a temporary vector to redefine the dimension of orderRecipe2
     *  if necessary (using  "p" or "c" steps) */ 
    public int[] orderRecipeTmp;

    /** the dictionary to be used to choose the steps to be included in 
        a recipe */
    public int[] dictionary;
    /** length of the dictionary */
    public int dictionaryLength;
 
    /** the tool used in assigning orders */
    AssigningTool assigningTool;

    // units
    Unit aUnit;
    EndUnit anEndUnit;

    /**
     * the constructor for OrderGenerator
     */
    public OrderGenerator (Zone aZone, int msn, int msl, ListImpl ul, 
			   ListImpl eul, ListImpl ol, 
			   VEFrameModelSwarm mo, AssigningTool at) 
    {

	// Call the constructor for the OrderGenerator parent class.
	super(aZone); int dim;

	orderCount=0;
	maxStepNumber=msn;
	if(maxStepNumber==0){
	    System.out.println("maxStepNumber cannot be 0 in this run.");
	    System.exit(1);}
	maxStepLength=msl;
	if(maxStepLength==0){
	    System.out.println("maxStepLength cannot be 0 this run.");
	    System.exit(1);}
	unitList=ul;
	endUnitList=eul;
        orderList=ol;

	orderRecipe1 = new int[maxStepNumber];
	stepLengthInOrder = new int[maxStepNumber];
	dim=maxStepNumber*maxStepLength; if(dim<10)dim=10;
	orderRecipe2 = new int[dim];

        vEFrameModelSwarm = mo;
	assigningTool=at;
    }

    /** building the dictionary containing the steps to be included in the
	recipes */
    public void setDictionary()
    {
	int i, ii, iii;
	// how many words in the dictionary?

	// starting point: the no. of endUnits 
	dictionaryLength=endUnitList.getCount(); 

	for (i=0;i<unitList.getCount();i++)
	    {
		aUnit=(Unit) unitList.atOffset(i);
		dictionaryLength += aUnit.getNOfPhasesToDealWith();
	    }   
	if(StartVEFrame.verbose)
	    System.out.println(
	 "Length of the dictionary (# of codes identifying steps in recipes): "
	 + dictionaryLength);

	//creating the dictionary
	dictionary = new int[dictionaryLength];

	// loading terms into the dictionary (from endUnits and then units)
	i=0;
	for (ii=0;ii<endUnitList.getCount();ii++)
	    {
		anEndUnit=(EndUnit) endUnitList.atOffset(ii);
		dictionary[i++]=anEndUnit.getUnitNumber();
	    }
	for (ii=0;ii<unitList.getCount();ii++)
	    {
		aUnit=(Unit) unitList.atOffset(ii);
		for (iii=0;iii<aUnit.getNOfPhasesToDealWith();iii++)
		    dictionary[i++]=aUnit.getProductionPhase(iii);
	    }

	if(StartVEFrame.verbose)
	    for (i=0;i<dictionaryLength;i++)
		System.out.println(i + "  " + dictionary[i]);
    }


    /** creating an order form, choosing its recipe length and content */
    public void createRandomOrderWithNSteps() {
	int i, ii, iii, k, count, kk, n, repeat, endRepeat, 
	    randomStepNumber, randomStepNumber0, randomStep;
	boolean finished, nextUnitIsNotAnEndUnit,
	    createAStandAloneBatchOrder, createASequentialBatchOrder,
	    createAnOrOrder;

	createAStandAloneBatchOrder=false;
	createASequentialBatchOrder=false;
	createAnOrOrder=false;
	orderCount++;
	count=0;

	if(unitList.getCount()==0){
	    System.out.println("totalUnitNumber cannot be 0 in this run.");
	    System.exit(1);}

	if(endUnitList.getCount()>0)
	    if(0.05>Globals.env.uniformDblRand.getDoubleWithMin$withMax(0,1))
		createAStandAloneBatchOrder=true;

	if(endUnitList.getCount()>0 && ! createAStandAloneBatchOrder) 
	    // per se sequential batches don't use
	    // end units; this is a switch avoinding
	    // this kind of order generation in case
	    // of plane use of the program 
	    
	    // anyway, the simple test recipes
	    // produced here in the sequential batch
	    // case are conclueded with an end unit
	    if(0.05>Globals.env.uniformDblRand.getDoubleWithMin$withMax(0,1))
		createASequentialBatchOrder=true;

	if(endUnitList.getCount()>0 && ! (createAStandAloneBatchOrder ||
					  createASequentialBatchOrder)) 
	    // per se 'or order' don't use necessarily	    
	    // end units; this is a switch avoinding
	    // this kind of order generation in case
	    // of plane use of the program 
	    if(0.05>Globals.env.uniformDblRand.getDoubleWithMin$withMax(0,1))
		createAnOrOrder=true;

	// ------------------------------
	// create a stand alone batch order
	if(createAStandAloneBatchOrder)
	    {
		// the recipe scheme of a stand alone batch is:

		// external recipe (with the time specification in seconds 
		// = ticks or in other time units)
		    
		// 11 s 10 / 5 e 101
		// which means: to produce a batch of 5 items in 10 ticks
		// in unit 11 and putting the result in e 101
		    
		// internally, in orderGenerator, with / reported as -2

		// NB NB in sequential batch "/" is substituted with "\"

		// -2 10 5 11 101

		// into anOrder

		// 11 101
		// with a standAloneBatchSpecificationSet object having 
		// the same # of the order and a position parameter relative 
		// to the recipe sequence (here position = 0)

		count=5;
		if(count>maxStepNumber){
		  System.out.println(
                  "OrderGenerator has to produce a stand alone batch order, "+
		  "so set maxStepNumber >= 5");
		  System.exit(1);
		}
		// code of stand alone batch processes
		orderRecipe2[0]=-2;
		// # of seconds = ticks of the whole process
		// (fixed only to simplify this version of the generator)
		orderRecipe2[1]=5;
		// # of items produced (fixed, see above)
		orderRecipe2[2]=10;
		// producing unit
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[3]=dictionary[n];
		// receiving end unit
		n = Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(0, endUnitList.getCount()-1);
		orderRecipe2[4]=dictionary[n];

		if(StartVEFrame.verbose){
		    System.out.println("Preparing order " 
				       + orderCount + " as a stand alone " +
				       "batch order, with recipe:");
		    for (n=0;n<5;n++)System.out.print(orderRecipe2[n] + " ");
		    System.out.println(" ");}
	    }


	// ------------------------------
	// create a sequential batch order
	if(createASequentialBatchOrder)
	    {
		// the recipe scheme of a stand alone batch is:

		// external recipe (with the time specification in seconds 
		// = ticks or in other time units)
		
		// 10 s 3 11 s 10 \ 5 12 s 2 e 101
		// which means: to produce using units 10, 11, 12 and
		// putting the result in e 101 (the final e is not 
		// mandatory) 

		// in unit 11 we have to wait for 5 order to make the step
		// globally in 10 ticks (NB NB here we use \)
		    
		// internally, in orderGenerator, with \ reported as -3

		// 10 10 10 -3 10 5 11 12 12 101

		// into anOrder

		// 10 10 10 11 12 12 101
		// with a sequentialBatchSpecificationSet object having the 
		// same # of the order and a position parameter relative 
		// to the recipe sequence (here position = 3)

		// the structure of these test recipes is fixed; the unit 
		// numbers are chosen randomly

		count=10;
		if(count>maxStepNumber){
		  System.out.println(
                  "OrderGenerator has to produce a sequential batch order, "+
		  "so set maxStepNumber >= 10");
		  System.exit(1);
		}
		// producing unit
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[0]=dictionary[n];
		orderRecipe2[1]=dictionary[n];
		orderRecipe2[2]=dictionary[n];

		// code of the sequential batch processes
		orderRecipe2[3]=-3;
		// # of seconds = ticks of the whole process
		// (used here as a constant only to simplify this 
		// version of the generator)
		orderRecipe2[4]=10;
		// # of items produced (as a costant, see above)
		orderRecipe2[5]=5;
		// producing unit (with the sequential batch)
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[6]=dictionary[n];
		// producing unit (with the sequential batch)
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[7]=dictionary[n];
		orderRecipe2[8]=dictionary[n];
		// receiving end unit
		n = Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(0, endUnitList.getCount()-1);
		orderRecipe2[9]=dictionary[n];

		if(StartVEFrame.verbose)
		    {
		    System.out.println("Preparing order " 
				       + orderCount + " as a sequential " +
				       "batch order, with recipe:");
		    for (n=0;n<10;n++)System.out.print(orderRecipe2[n] + " ");
		    System.out.println(" ");
		    }
	    }


	// ------------------------------
	// create an 'or order' (i.e an order containing an or operator)
	if(createAnOrOrder)
	    {
		// the recipe scheme of an 'or order' is:

		// external recipe (with the time specification in seconds 
		// = ticks or in other time units; to check the code, in the
		// the first branch of the 'or' we have procurement process
		// with 101 below as an end unit)
		
		// 10 s 3 || 1 11 s 2 p 1 101 10 s 1 9 s 2 
		// || 2 14 s 4 || 0 6 s 2
		// where || 1 and || 2 are to nodes of or sequences (any 
		// number) and || 0 concludes them

		// in time sequence the recipe is

		// 10 s 3 || 1 11 s 2 p 1 101 10 s 1 9 s 2 || 0 6 s 2
		//        || 2 14 s 4              
		// with || 1 and || 2 branches mutually exclusive

		// internally, in orderGenerator, with || reported as -10

		// 10 10 10 -10 1 11 11 -1 1 1 10 9 9 -10 2 14 14 14 14 
		// -10 0 6 6

		// into anOrder

		// 10 10 10       11 11 10 9 9       14 14 14 14       6 6
		// with an OrSpecificationSet objects having the 
		// same # of the order and all the position parameters 
		// relative to the 'or' sequence in the recipe 
		// (here position = 3, 8, 12; first position is 0)
		// with a p specification (at position 5)

		// the structure of these test recipes is fixed; the unit 
		// numbers are chosen randomly

		count=23;
		if(count>maxStepNumber){
		  System.out.println(
                  "OrderGenerator has to produce an 'or' order, "+
		  "so set maxStepNumber >= 20");
		  System.exit(1);
		}
		// producing unit
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[0]=dictionary[n];
		orderRecipe2[1]=dictionary[n];
		orderRecipe2[2]=dictionary[n];
		// codes of the or process
		orderRecipe2[3]=-10;
		orderRecipe2[4]= 1;
		// producing unit
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[5]=dictionary[n];
		orderRecipe2[6]=dictionary[n];
		// procurement
		orderRecipe2[7]=-1;
		orderRecipe2[8]=1;
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(0,endUnitList.getCount()-1);
		orderRecipe2[9]=n;
		// producing unit
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[10]=dictionary[n];
		// producing unit
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[11]=dictionary[n];
		orderRecipe2[12]=dictionary[n];
		// codes of the or process
		orderRecipe2[13]=-10;
		orderRecipe2[14]= 2;
		// producing unit
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[15]=dictionary[n];
		orderRecipe2[16]=dictionary[n];
		orderRecipe2[17]=dictionary[n];
		orderRecipe2[18]=dictionary[n];
		// codes of the or process
		orderRecipe2[19]=-10;
		orderRecipe2[20]= 0;
		// producing unit
		n= Globals.env.uniformIntRand.
		    getIntegerWithMin$withMax(endUnitList.getCount(),
					      dictionaryLength-1);
		orderRecipe2[21]=dictionary[n];
		orderRecipe2[22]=dictionary[n];

		if(StartVEFrame.verbose)
		    {
			System.out.println("Preparing order " 
					   + orderCount + " as an Or " +
					   "order, with recipe:");
			for (n=0;n<count;n++)System.out.print(
					     orderRecipe2[n] + " ");
			System.out.println(" ");
		    }
	    }

	// ------------------------------
	// normal form creation process
	if(! (createAStandAloneBatchOrder || createASequentialBatchOrder || 
	      createAnOrOrder))
	{
	randomStepNumber=Globals.env.uniformIntRand.getIntegerWithMin$withMax
		    (1, maxStepNumber);

	// generating the sequence in the recipe

	// putting 0 in the vector may be useful for future software 
	// modifications
	for (i=0;i<maxStepNumber;i++)orderRecipe1[i]=0;
	for (i=0;i<maxStepNumber;i++)stepLengthInOrder[i]=0;

	finished=false;
	randomStepNumber0=0;

	for (i=0;i<randomStepNumber && ! finished;i++)
	    {
		randomStepNumber0=i+1;

		// generating a random step and stating the probability
		// of concluding the recipe with an "e" end code
		if(0.10 >= Globals.env.uniformDblRand.
		   getDoubleWithMin$withMax(0,1) && 
		   endUnitList.getCount()>0)		
		    randomStep = Globals.env.uniformIntRand.
			getIntegerWithMin$withMax(0, endUnitList.getCount()-1);
		else
		    randomStep = Globals.env.uniformIntRand.
			getIntegerWithMin$withMax(endUnitList.getCount(),
						  dictionaryLength-1);
		orderRecipe1[i]= dictionary[randomStep];

		// check if an end code has been generated
		if (randomStep < endUnitList.getCount())finished=true;

		// generating the length of each step
		if (maxStepLength>1 && ! finished) // end step has lenght 1
		    stepLengthInOrder[i]=
			Globals.env.uniformIntRand.getIntegerWithMin$withMax
			(1, maxStepLength);
		else 
		    stepLengthInOrder[i]=1;
	    }

	// putting together
	// (i) sequences "p" or "c" or ... 
	// (ii) and the lengths of the steps
	count=0;
	for (i=0;i<randomStepNumber0;i++)
	    for (ii=0;ii<stepLengthInOrder[i];ii++)
		{

		    // introducing "p" or "c" sequences

		    // the "p" strucuture is
		    // p k a1 a2 .... ak
		    // where k is the # of items to be procured and 
		    // a1, a2, ..., ak are the codes of those items 
		    // (corresponding to "e" or endUnit codes in sub-recipes
		    // related to internal or external sub-processes

		    // the recipe scheme is
		    // external recipe (with the time specifications and 
		    // a tick corresponding to 1 sec. or to an analogous
		    // minimum fraction of the time
		    
		    // 11 s 3 p 3 101 102 103 22 s 2
		    
		    // internally, in orderGenerator, with p reported as -1

		    // 11 11 11 -1 3 101 102 103 22 22

		    // into anOrder

		    // 11 11 11 22 22
		    // with a procurementSpecifications object having the same 
		    // # of the order and a position parameter relative to the
		    // recipe sequence (here position = 3, starting from 0)

		    
		    if(endUnitList.getCount() > 0)
			{
			    nextUnitIsNotAnEndUnit=true;
			    for(kk=0;kk<endUnitList.getCount();kk++)
				if(orderRecipe1[i]==((EndUnit) endUnitList.
						 atOffset(kk)).getUnitNumber())
				    nextUnitIsNotAnEndUnit=false;
			    
			    // conditions: first item of a step (ii==0)
			    //             next unit number != endUnit number
			    //             (nextUnitIsNotAnEndUnit)
			    if (ii==0 && nextUnitIsNotAnEndUnit &&
				0.10 >= Globals.env.uniformDblRand.
				getDoubleWithMin$withMax(0,1)){

				// length of the "p" or "c" step
				k=Globals.env.uniformIntRand.
				    getIntegerWithMin$withMax
				    (1, endUnitList.getCount());

				// increase the length of orderRecipe2 vector
				orderRecipeTmp = new int[orderRecipe2.
							length+2+k];
				System.arraycopy(orderRecipe2,0,
						 orderRecipeTmp,0,
						 orderRecipe2.length);
				orderRecipe2=orderRecipeTmp;

				//the code of process p
				orderRecipe2[count++]= -1; 

				// the length
				orderRecipe2[count++]=k;   
				
				// k codes, chosen randomly
				for (iii=0;iii<k;iii++)   
				    orderRecipe2[count++]=
					((EndUnit) endUnitList.
					 atOffset(
						  Globals.env.uniformIntRand.
						  getIntegerWithMin$withMax
						  (0, endUnitList.getCount()-1)
						  )).getUnitNumber();
			    }
			}

		    // copying normal steps
		    orderRecipe2[count++]=orderRecipe1[i];

		}
	
	if(StartVEFrame.verbose)
	{
	    for (iii=0; iii<count;iii++)
		System.out.print(orderRecipe2[iii]+ " ");
	    System.out.println(" internal recipe for order " + orderCount);
	}

	// ------------------------------
	// concluding normal order form generation
        }


	// repeating the order generation (in sequential batch order case)
	endRepeat=1;

	// deliberately the repetitions here are five, being also the
	// the whole batch composed of five order; due to the delay in 
	// assigning order to the unit using the sequential batch process,
	// the sequential batch assembler cannot anyway clear the batch 
	// for production at the first check (this delay in clearing is 
	// useful for testing purposes)
	if(createASequentialBatchOrder)endRepeat=5;

	for(repeat=0;repeat<endRepeat;repeat++){

	// creating an order
        // vector dimension in anOrder is count: count is a step ahead (see 
        // last count++ operation) but starting with 0 is the correct number
        // of elements
	anOrder = new Order(getZone(), orderCount,
			    Globals.env.getCurrentTime(),
			    count, orderRecipe2, 
			    vEFrameModelSwarm, endUnitList);

	if(StartVEFrame.verbose)
	{
	    for (iii=0; iii<anOrder.orderRecipe.length;iii++)
		System.out.print(anOrder.orderRecipe[iii]+ " ");
	    System.out.println(" internal recipe into the order " +
			       orderCount);
	}


        // add the active orders to the general order list (they will be
	// eliminated when dropped in a unit, being finished); this
	// list has been introduced for accounting purposes [may be it would
	// be better subtitute it with a get to the units to know their
	// waiting lists] 
        orderList.addLast(anOrder);

	/**
	 * sending the order to the first production unit
	 * (we are acting as the Front End of the VE)
	 */
	assigningTool.assign(anOrder);
/*	unitNotFound=true;
	for (i=0;i<unitList.getCount() && unitNotFound;i++)
	    {
		aUnit=(Unit) unitList.atOffset(i);
		for (iii=0;iii<aUnit.getNOfPhasesToDealWith() &&  
			 unitNotFound;iii++)
		    {
		       if(anOrder.getNextStep()==aUnit.getProductionPhase(iii))
			    {
				unitNotFound=false;
				aUnit.setWaitingList(anOrder);
			    }
		    }
	    }

	// considering also endUnits
	for (i=0;i<endUnitList.getCount() && unitNotFound;i++)
	    {
		anEndUnit=(EndUnit) endUnitList.atOffset(i);
		if(anOrder.getNextStep()==anEndUnit.getUnitNumber())
		    {
			unitNotFound=false;
			anEndUnit.setTemporaryList(anOrder);
		    }
	    }



	if(unitNotFound){
	    System.out.println("OrderGenerator internal error: receiving "+
			       "unit does not exixts");
	    System.exit(1);
	}
*/
	// finish repeating
	if(endRepeat>1 && repeat<4)orderCount++;
	}

    }

}
